import math
import numpy

def guess_fuzzy_centroids(problem_space, nclusters):
    min = problem_space[0][0]
    max = problem_space[-1][0]
    range = float(max - min)
    distance = range/nclusters
    offset = distance/2.0
    centroids = [min + offset + distance * c for c in xrange(nclusters)]
    return numpy.array(centroids)

def normalize_fuzzy_memberships(clusters):
    clusters = clusters.T
    clusters = numpy.array([cluster/sum(cluster) for cluster in clusters])
    clusters = clusters.T
    return clusters

def calculate_fuzzy_partition(problem_space, fuzzy_centroids, m = 2.0):
    nelements = len(problem_space)
    nclusters = len(fuzzy_centroids)
    fuzzy_partition = [[None] * nelements for j in xrange(nclusters)]
    for i in xrange(nelements):
        for j in xrange(nclusters):
            sigma = 0.0
            for k in xrange(nclusters):
                numer = problem_space[i][0] - fuzzy_centroids[j]
                denom = problem_space[i][0] - fuzzy_centroids[k]
                sigma += (numer / denom) ** (2.0 / (m - 1.0))
                fuzzy_partition[j][i] = 1.0 / sigma
            if math.isnan(fuzzy_partition[j][i]): fuzzy_partition[j][i] = 1.0
    return normalize_fuzzy_memberships(numpy.array(fuzzy_partition))

def calculate_fuzzy_centroids(problem_space, solution_space, m = 2.0):
    nclusters = len(solution_space)
    nelements = len(problem_space)
    ret_vec = [None] * nclusters
    for j in xrange(nclusters):
        numer = sum([solution_space[j,i] ** m * problem_space[i][1] * problem_space[i][0] for i in xrange(nelements)])
        denom = sum([solution_space[j,i] ** m * problem_space[i][1] for i in xrange(nelements)])
        ret_vec[j] = numer/denom
    return numpy.array(ret_vec)

def calculate_epsilon(previous_centroids, new_centroids):
    return sum([(p - n) ** 2.0 for p, n in zip(previous_centroids, new_centroids)]) ** 0.5

def defuzzify_fuzzy_partition(problem_space, fuzzy_partition, fuzzy_centroids):
    fuzzy_centroids = [int(i) for i in fuzzy_centroids]
    retObj = {}
    for i in xrange(len(problem_space)):
        winning_value = fuzzy_partition[0][i]
        winning_centroid = 0
        for j in xrange(len(fuzzy_centroids)):
            if fuzzy_partition[j][i] > winning_value:
                winning_value = fuzzy_partition[j][i]
                winning_centroid = j
        retObj[problem_space[i][0]] = fuzzy_centroids[winning_centroid]
    return retObj

def defuzzify_image(defuzz_dict, image_data):
    retObj = [[None] * len(image_data[0]) for i in xrange(len(image_data))]
    for r in xrange(len(image_data)):
        for c in xrange(len(image_data[0])):
            retObj[r][c] = defuzz_dict[image_data[r][c]]
    return numpy.array(retObj)

def FCM(problem_space, nclusters, m = 2.0, quitting_epsilon = 0.001, maxiters = 100, statistics = False, verbose = False):
    fuzzy_centroids = guess_fuzzy_centroids(problem_space, nclusters)
    fuzzy_partition = None
    prev_fuzzy_centroids = None
    for i in xrange(maxiters):
        fuzzy_partition = calculate_fuzzy_partition(problem_space, fuzzy_centroids)
        fuzzy_centroids = calculate_fuzzy_centroids(problem_space, fuzzy_partition)
        if verbose and i % 10 == 0:
            print i, 0, XB_partition_fitness(problem_space, fuzzy_partition, fuzzy_centroids), fuzzy_centroids
        if prev_fuzzy_centroids != None:
            centroid_epsilon = calculate_epsilon(prev_fuzzy_centroids, fuzzy_centroids)
            if centroid_epsilon < quitting_epsilon:
                break
        prev_fuzzy_centroids = fuzzy_centroids
    if verbose and i % 10 != 0:
        print i, 0, XB_partition_fitness(problem_space, fuzzy_partition, fuzzy_centroids), fuzzy_centroids
    defuzz_dict = defuzzify_fuzzy_partition(problem_space, fuzzy_partition, fuzzy_centroids)
    if statistics == False:
        return defuzz_dict
    else:
        return defuzz_dict, fuzzy_partition, fuzzy_centroids

def XB_partition_fitness(problem_space, fuzzy_partition, centroids):
    min_centroid_eudist = abs(centroids[0] - centroids[1])
    for i in xrange(len(centroids)):
        for j in xrange(i+1, len(centroids)):
            this_dist = abs(centroids[j] - centroids[i])
            if this_dist < min_centroid_eudist:
                min_centroid_eudist = this_dist
    numer = 0.0
    for j in xrange(len(centroids)):
        for i in xrange(len(problem_space)):
            numer += fuzzy_partition[j][i] ** 2.0 * problem_space[i][1] * (problem_space[i][0] - centroids[j]) ** 2.0
    denom = sum([ps[1] for ps in problem_space]) * min_centroid_eudist ** 2.0
    # WHY SQUARE THE EUCLIDEAN DISTANCE?!
    return numer / denom

def PC_partition_fitness(fuzzy_partition, problem_space = None):
    sigma = 0.0
    for j in xrange(len(fuzzy_partition)):
        for i in xrange(len(fuzzy_partition[0])):
            if problem_space != None:
                sigma += fuzzy_partition[j][i] ** 2.0 * problem_space[i][1]
            else:
                sigma += fuzzy_partition[j][i] ** 2.0
    return sigma/len(fuzzy_partition[0])

def PE_partition_fitness(fuzzy_partition, problem_space = None):
    sigma = 0.0
    for j in xrange(len(fuzzy_partition)):
        for i in xrange(len(fuzzy_partition[0])):
            if problem_space != None:
                sigma += fuzzy_partition[j][i] * math.log(fuzzy_partition[j][i]) * problem_space[i][1]
            else:
                sigma += fuzzy_partition[j][i] * math.log(fuzzy_partition[j][i])
    return -sigma/len(fuzzy_partition[0])

def PBMF_partition_fitness(problem_space, partition, centroids, E1 = 5000.0, p = 2.0):
    factor1 = len(centroids) ** -1.0
    factor2numer = E1
    factor2denom = 0.0
    factor3 = 0.0
    for j in xrange(len(partition)):
        for i in xrange(len(partition[0])):
            factor2denom += partition[j][i] * abs(problem_space[i][0] - centroids[j])
    for a in xrange(len(centroids)):
        for b in xrange(a, len(centroids)):
            this_dist = abs(centroids[a] - centroids[b])
            if this_dist > factor3:
                factor3 = this_dist
    return (factor1 * factor2numer / factor2denom * factor3) ** p
